#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include "bluetooth/bluetooth.h"
#include "bluetooth/hci.h"
#include "bluetooth/hci_lib.h"
#include "bluetooth/sdp.h"
#include "bluetooth/sdp_lib.h"

#include "btconstants.h"
#include "btfunctions.h"

int main(int argc, char **argv) 
{
    uint32_t 				range = 0x0000ffff;
    struct hci_dev_info 	di;
    struct hci_version 		version;
	struct hci_conn_info_req *cr;
	uint16_t 				handle;
	inquiry_info 			*ii = NULL;
	int 					max_rsp, num_rsp;
	int 					dev_id, sock, len, flags;
	int 					iIter;
	char 					addr[20] = { 0 };
	char 					name[248] = { 0 };
	struct search_context 	context;
	sdp_list_t 				*attrid, *search, *seq, *next;
	sdp_session_t 			*sess;
	sdp_record_t 			*rec;
	sdp_list_t				*list;
	sdp_data_t 				*sdpdata;
	uint16_t 				attrId;
	unsigned char			handle_ok;
	uint8_t 				features[8];
	int						channel=-1;
	bp_sdp_t	 			*bp_list=NULL;
	
	
	printf("Content-type: text/html\n\n<html><body>");
	dev_id = hci_get_route(NULL);
	sock = hci_open_dev( dev_id );
	if (dev_id < 0 || sock < 0) {
		perror("opening socket failed!");
		exit(1);
	}
	printf("Using dev_id: %i\n<br>",dev_id);
	if(hci_devinfo(0, &di) < 0) {
		perror("HCI device info failed");
		exit(1);
	}
	ba2str(&(di.bdaddr), addr);
	printf("Got dev info. My MAC: %s\n<br>",addr);
	
	len = 4;
	max_rsp = 10;
	flags = IREQ_CACHE_FLUSH;
	ii = (inquiry_info*)malloc(max_rsp * sizeof(inquiry_info));
	
	printf("Inquiring...\n<br>");
	num_rsp = hci_inquiry(dev_id, len, max_rsp, NULL, &ii, flags);
	if ( num_rsp < 0) perror("hci_inquiry failed");
	printf("Inquiry finished. Found %i devices.\n<br>",num_rsp);
		
	for (iIter=0; iIter<num_rsp; iIter++) {
		ba2str(&(ii+iIter)->bdaddr, addr);
		memset(name, 0, sizeof(name));

		/**************************************************************
		 * 
		 * HCI Blueprinting for each found device
		 * 
		 **************************************************************/

		handle_ok = 0;
		cr = malloc(sizeof(*cr) + sizeof(struct hci_conn_info));
		if (!cr) {
			printf("Allocating memory for struct cr failed!\n");
		} else {
			bacpy(&cr->bdaddr, &(ii+iIter)->bdaddr);
			cr->type = ACL_LINK;
			if (ioctl(sock, HCIGETCONNINFO, (unsigned long) cr) < 0) {
				if (hci_create_connection(sock, &(ii+iIter)->bdaddr,htobs(di.pkt_type & ACL_PTYPE_MASK),0, 0x01, &handle, 25000) < 0) {
					perror("Can't create connection");
				} else {
					handle_ok = 1;
				}
			} else {
				handle = htobs(cr->conn_info->handle);
				handle_ok = 1;
			}
			if (handle_ok == 1) {
				if (hci_read_remote_name(sock, &(ii+iIter)->bdaddr, sizeof(name), name,100000) < 0) 
					strcpy(name,"[unknown]");
				memset(&version, '\0', sizeof(struct hci_version));
				hci_read_remote_version(sock, handle, &version, 20000);
				printf("<hr>****************************************************************<br>\n");
				printf("<b>%s: %s</b><br><br>\n",addr,name);
				/*
				printf("LMP version: 0x%X.0x%X\n",version.lmp_ver,version.lmp_subver);
				printf("HCI version: 0x%X.0x%X\n",version.hci_ver,version.hci_rev);
				printf("Manufacture: 0x%X\n",version.manufacturer);
				*/

				printf("<form method=\"POST\" action=\"/bbt/action.add.php\">\n");
				printf("Default name of the device: <input type=\"text\" name=\"defaultname\" value=\"%s\" size=\"20\"><br><br>",name);
				
				hci_read_remote_features(sock, handle, features, 20000);
				//printf("Features   : 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X\n",features[0],features[1],features[2],features[3],features[4],features[5],features[6],features[7]);
				
				print_hci_id_html(&(ii+iIter)->bdaddr,&version,features);
			} 
			free(cr);
		}
		
		
		
		
		
		/**************************************************************
		 * 
		 * SDP Blueprinting for each found device
		 * 
		 **************************************************************/
		
		
		
		//printf ("SDP inquiry: \n");
		/* Zero out the search context */
		memset(&context, '\0', sizeof(struct search_context));
		sdp_uuid16_create(&context.group, PUBLIC_BROWSE_GROUP);
		sess = sdp_connect(&di.bdaddr, &(ii+iIter)->bdaddr, SDP_RETRY_IF_BUSY);
		if(!sess) {
			perror("Failed to connect to SDP server.");
		} else {
			attrid = sdp_list_append(0, &range);
			search = sdp_list_append(0, &context.group);
			if (sdp_service_search_attr_req(sess, search, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
				printf("Service Search failed.\n");
				sdp_close(sess);
				return -1;
			}
			sdp_list_free(attrid, 0);
			sdp_list_free(search, 0);
			//printf("SDP result: \n");
			for (; seq; seq = next) {
				rec = (sdp_record_t *) seq->data;
				list = (sdp_list_t *)rec->attrlist;
				for (; list; list = list->next) {
					sdpdata = (sdp_data_t *)list->data;
					attrId = sdpdata->attrId;
					/*
					 * sdp_get_attribs_from_record reads attribs from record
					 * and stores it in an own struct bp_sdp (bpconstants.h).
					 * This struct is a list and it will be finally used
					 * to generate the SDP print
					 */
					bp_list = sdp_get_attribs_from_record(sdpdata,attrId,bp_list);
				}
				if (seq->next) {
					bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,0xFF,SDP_SPACER,NULL,0);
				}
				next = seq->next;
				free(seq);
				sdp_record_free(rec);
			}
			print_sdp_id_html(bp_list,"serv_");
			//printf("Freeing SDP list\n");
			free_bp_list(bp_list);
			bp_list = NULL;
			sdp_close(sess);
			
			
			
			
			
			/**************************************************************
			 * 
			 * SDP Blueprinting only for OBEX Oush service
			 * 
			 **************************************************************/
			memset(&context, '\0', sizeof(struct search_context));
			sdp_uuid16_create(&context.group, OBEX_OBJPUSH_SVCLASS_ID);
			sess = sdp_connect(&di.bdaddr, &(ii+iIter)->bdaddr, SDP_RETRY_IF_BUSY);
			if(!sess) {
				perror("Failed to connect to SDP server.");
			} else {
				attrid = sdp_list_append(0, &range);
				search = sdp_list_append(0, &context.group);
				if (sdp_service_search_attr_req(sess, search, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
					printf("Service Search failed.\n");
					sdp_close(sess);
					return -1;
				}
				sdp_list_free(attrid, 0);
				sdp_list_free(search, 0);
				for (; seq; seq = next) {
					rec = (sdp_record_t *) seq->data;
					list = (sdp_list_t *)rec->attrlist;
					for (; list; list = list->next) {
						sdpdata = (sdp_data_t *)list->data;
						attrId = sdpdata->attrId;
						/*
						 *  Evaluate only the ProtocolDescriptorList (attrId == 0x04)
						 */
						if (attrId == 0x04) {
							sdp_get_opush_channel(sdpdata,attrId,-1,0,&channel);
						}
						bp_list = sdp_get_attribs_from_record(sdpdata,attrId,bp_list);
					}
					if (seq->next) {
						bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,0xFF,SDP_SPACER,NULL,0);
					}
					next = seq->next;
					free(seq);
					sdp_record_free(rec);
				}
				print_sdp_id_html(bp_list,"opush_");
				printf("<br>OPUSH Channel: <input type=\"text\" name=\"opushchannel\" value=\"%i\"><br>\n\n",channel);
				free_bp_list(bp_list);
				bp_list = NULL;
				sdp_close(sess);
			}

			
			
			printf("\n<br><br><input type=\"submit\" value=\"Save this device\">");
			printf("</form>");
		}
		
		
	}
	printf("</body></html>");
	
	free(ii);
	close(sock);
	return 0;
}
