#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/sem.h>
#include <sys/ipc.h>
#include "bluetooth/bluetooth.h"
#include "bluetooth/hci.h"
#include "bluetooth/hci_lib.h"
#include "bluetooth/sdp.h"
#include "bluetooth/sdp_lib.h"

#include "btconstants.h"
#include "btfunctions.h"

int main(int argc, char **argv) 
{
	mydevices_t				*mydevs=NULL, *mycurdev=NULL;
	founddevices_t			**remotedevs;
	lockeddevices_t			**lockeddevs;
	thread_param_t			*params;
	int						devices_avail = 0, devices_ok=0, sock=0;
	struct sembuf 			sb;
	
	remotedevs = malloc(sizeof(founddevices_t*));
	lockeddevs = malloc(sizeof(lockeddevices_t*));
	*remotedevs = NULL;
	*lockeddevs = NULL;
	
	
	/*
	bdaddr_t a,b;
	add_locked_device(lockeddevs,&a);
	add_locked_device(lockeddevs,&b);
	dump_locked_devices(lockeddevs);
	return 0;
	*/
	
	/*
	 * Find my devices 
	 */
	for (devices_avail=0; devices_avail<DEVICES_MAX; devices_avail++) {
		sock = hci_open_dev( devices_avail );
		if (sock >= 0) {
			close(sock);
			mydevs = add_local_device(mydevs, devices_avail, PURPOSE_SCAN);
			devices_ok++;
		}
	}

	dump_local_devices(mydevs);

	/*
	 * Generate semaphores
	 */
	semkey = ftok("bbt-runner", 'S');
	semid	= semget(semkey, 1, 0666 | IPC_CREAT);
	if (semid < 0) {
		perror("Creating semaphore.");
	}
	sb.sem_num 	= SEM_SCAN;
	sb.sem_op	= 1;
	sb.sem_flg	= SEM_UNDO;
	semop(semid, &sb, 1);

	sb.sem_num 	= SEM_LOCKS;
	sb.sem_op	= 1;
	sb.sem_flg	= SEM_UNDO;
	semop(semid, &sb, 1);

	/*
	 *  Thread management
	 */	
	
	while (1) {

		/*********************************
		 * 	HCI SCAN
		 *********************************/
		// Starting scanner threads
		mycurdev = get_free_local_device(mydevs, PURPOSE_SCAN);
		if (mycurdev != NULL) {
			params = malloc(sizeof(thread_param_t));
			params->local_device_list 	= mydevs;
			params->remote_device_list 	= (void**)remotedevs;
			params->locked_devices		= (void**)lockeddevs;
			params->use_this_device 	= mycurdev;
			params->use_this_remote_device = NULL;			
			mycurdev->status			= STATUS_PROCESSING;
			mycurdev->tstamp			= time(NULL);
			pthread_create(&(mycurdev->thread_id), NULL, thread_scan, params);
			pthread_detach(mycurdev->thread_id);
		} else {
			// printf("No free scan device!\n");
		}

		mycurdev = get_free_local_device(mydevs, PURPOSE_PUSH);


		/*********************************
		 * 	Cleanup process
		 *********************************/
		cleanup_expired_remote_devices(remotedevs, lockeddevs, DEVICE_RTIMEOUT);
//		dump_remote_devices(remotedevs);
		
		
		sleep(DEVICE_TICKRATE);
	}
	
	/*
	 * Removing semaphores
	 */
	semctl(semid, SEM_SCAN, IPC_RMID, 0);
	semctl(semid, SEM_LOCKS, IPC_RMID, 0);

	// Cleanup all remote devices by setting time_gap to -1
	printf("Cleaning up expired\n");
	cleanup_expired_remote_devices(remotedevs, lockeddevs, -1);	
	printf("Cleaning up locked\n");	
	free_locked_devices(lockeddevs, *lockeddevs);
	printf("Cleaning up local\n");
	free_local_devices(mydevs);
	
	free(lockeddevs);
	free(remotedevs);
	
	return 0;
}
